/*Lock是jdk1.5之后出现的用来取代synchronized
1、lock需要显式的加锁与释放锁: lock.lock(),与lock.unlock()
2、Lock支持多个相关联的锁对象,在线程同步中的线程间通信时，synchronized为避免线程全部等待
	需要使用notifyAll方法，降低执行效率。而lock可以创建多个condition（取代synchronized中唯一的Object），
	从而可以明确的唤醒对方线程，condition.signal()
3、如果使用synchronized不使用notifyAll，使用多个锁，容易出现死锁
程序思路：此程序模仿生产者--消费者例子
	一、多线程间通讯是多个线程操作同一份数据而执行的动作不同，为了满足一些需求（如：执行的先后），使用
	wait、notify或者await、signal等协调线程间的执行
	二、线程的等待和唤醒需要配合标志使用flag，线程被唤醒后仍需判断标志，使用while而不是if

*/
import java.util.concurrent.locks.*;
public class Test9{
	public static void main(String[] args){
		Demo d = new Demo();
		DemoIn in = new DemoIn(d);//线程同步执行不同操作的类
		DemoOut out = new DemoOut(d);
		//同时开启多个线程
		new Thread(in).start();
		new Thread(out).start();
		new Thread(in).start();
		new Thread(out).start();
		new Thread(in).start();
		new Thread(out).start();
	}
}
class Demo {
	private int x = 1;
	private boolean flag = false;
	final Lock lock = new ReentrantLock();//创建锁对象
    final Condition con_in  = lock.newCondition(); //同一个锁对应两个不同的线程等待，唤醒对象
    final Condition con_out = lock.newCondition(); 
	public void in(){
		lock.lock();
			while(flag)//使用while而不是if，线程被唤醒后，仍需判断标记，否则出现线程执行顺序错乱
				try{
					con_in.await();//让本线程等待
					System.out.println(Thread.currentThread().getName()+"---"+x);
					x=x+1;
					flag=true;
					con_out.signal();//明确唤醒对方线程
				}catch(InterruptedException e){}
				finally{
					lock.unlock();//如果发生异常也必须释放锁
				}	
	}
	public void out(){
		lock.lock();
		while(!flag)
			try{
				con_out.await();
				System.out.println(Thread.currentThread().getName()+"***"+x);
				flag = false;
				con_in.signal();
			}catch(InterruptedException e){}
			finally{
				lock.unlock();
			}
	}
}
/*class Demo {
	private int x = 1;
	private boolean flag = false;
	public synchronized void in(){//同步函数
			while(flag)
				try{this.wait();}catch(InterruptedException e){}
			System.out.println(Thread.currentThread().getName()+"---"+x);
			x=x+1;
			flag=true;
			this.notifyAll();//唤醒所有线程，否则可能出现所有线程等待

	}
	public synchronized void out(){

		while(!flag)
			try{this.wait();}catch(InterruptedException e){}
		System.out.println(Thread.currentThread().getName()+"***"+x);
		flag=false;
		this.notifyAll();
	}
}*/
class DemoIn implements Runnable{
	Demo d;
	DemoIn(Demo d){
		this.d = d;
	}
	public void run(){
		while(true){
			d.in();	
		}
	}
}

class DemoOut implements Runnable{
	Demo d;
	DemoOut(Demo d){
		this.d = d;
	}
	public void run(){
		while(true){
			d.out();
		}
	}
}